#include "avswrapper.h"
#include "tools.h"
#include "main.h"
#include "config.h"

#include <windows.h>
#include <stdio.h>

#include "safefgets/safefgets.h"

AVSWrapper::AVSWrapper()
{
	g_IsAnalyzing++;
	m_ScriptEnv = NULL;
	m_Clip      = NULL;
	m_NextFrame = 0;
	m_DllInstance = LoadLibrary(TEXT("AVISYNTH.DLL"));
	if (m_DllInstance) {
		IScriptEnvironment * (__stdcall * CreateEnv) (int);
		CreateEnv = (IScriptEnvironment * (__stdcall *) (int)) 
				GetProcAddress(m_DllInstance, "CreateScriptEnvironment"); 
		if (CreateEnv) {
			m_ScriptEnv = CreateEnv(AVISYNTH_INTERFACE_VERSION);
		}
	} 
}

AVSWrapper::~AVSWrapper()
{
	m_Clip = NULL;
	delete m_ScriptEnv;
	if (m_DllInstance) FreeLibrary(m_DllInstance);
	g_IsAnalyzing--;
}

bool AVSWrapper::IsOK()
{
	return (m_ScriptEnv != NULL);
}

bool AVSWrapper::Open(const String & fileName, const String & opener, bool upsideDown, const String & preProcessor, bool pal)
{
	if (! IsOK()) return false;
	Config::FunctionMap::const_iterator openerIterator = g_Config->GetOpeners().find(opener);
	String openerImport = (openerIterator == g_Config->GetOpeners().end() ? S("") : (*openerIterator).second);
	Config::FunctionMap::const_iterator preProcessorIterator = g_Config->GetPreProcessors().find(preProcessor);
	String preProcessorImport = (preProcessorIterator == g_Config->GetPreProcessors().end() ? S("") : (*preProcessorIterator).second);
	if (openerImport == preProcessorImport) preProcessorImport.clear();
	try {
		if (openerImport.length()) {
			char * s = new char[openerImport.length() + 1];
			strcpy(s, openerImport.toAnsi().c_str());
			AVSValue p = s;
			m_ScriptEnv->Invoke("Import", AVSValue(& p, 1));
			delete [] s;
		}
		if (preProcessorImport.length()) {
			char * s = new char[preProcessorImport.length() + 1];
			strcpy(s, preProcessorImport.toAnsi().c_str());
			AVSValue p = s;
			m_ScriptEnv->Invoke("Import", AVSValue(& p, 1));
			delete [] s;
		}
		char * fn = new char[fileName.length() + 1];
		strcpy(fn, fileName.toAnsi().c_str());
		AVSValue p[2];
		p[0] = fn;
		p[1] = pal;
		if (opener.length()) {
			AVSValue clip = m_ScriptEnv->Invoke(opener.toAnsi().c_str(), AVSValue(p, 2));
			m_Clip = (clip.IsClip() ? clip.AsClip() : NULL);
		} else {
			AVSValue clip = m_ScriptEnv->Invoke("Import", AVSValue(p, 1));
			m_Clip = (clip.IsClip() ? clip.AsClip() : NULL);
		}
		delete [] fn;
		if (upsideDown) {
			p[0] = m_Clip;
			AVSValue clip = m_ScriptEnv->Invoke("FlipVertical", AVSValue(p, 1));
			m_Clip = (clip.IsClip() ? clip.AsClip() : NULL);
		}
		if (m_Clip && preProcessor.length()) {
			p[0] = m_Clip;
			AVSValue clip = m_ScriptEnv->Invoke(preProcessor.toAnsi().c_str(), AVSValue(p, 2));
			m_Clip = (clip.IsClip() ? clip.AsClip() : NULL);
		}
	} catch (AvisynthError & error) {
		m_Error = String().fromAnsi(error.msg).firstLine();
		m_Clip = NULL;
	} catch (...) {
		m_Clip = NULL;
	}
	return RenderInfo();
}

bool AVSWrapper::ImageSource(const String & fileName, int width, int height)
{
	if (! IsOK()) return false;
	char * fn = new char [fileName.length() + 1];
	strcpy(fn, fileName.toAnsi().c_str());
	const char * argName[3];
	AVSValue argValue[3];
	int argCount = 0;
	argName[argCount] = "file";
	argValue[argCount++] = fn;
	argName[argCount] = "start";
	argValue[argCount++] = 0;
	argName[argCount] = "end";
	argValue[argCount++] = 0;
	try {
		AVSValue clip = m_ScriptEnv->Invoke("ImageSource", AVSValue(argValue, argCount), argName);
		if (clip.IsClip() && width && height) {
			argCount = 0;
			argValue[argCount++] = clip;
			argValue[argCount++] = width;
			argValue[argCount++] = height;
			clip = m_ScriptEnv->Invoke("Lanczos4Resize", AVSValue(argValue, argCount));
		}
		m_Clip = (clip.IsClip() ? clip.AsClip() : NULL);
	} catch (AvisynthError & error) {
		m_Error = String().fromAnsi(error.msg).firstLine();
		m_Clip = NULL;
	} catch (...) {
		m_Clip = NULL;
	}
	delete [] fn;
	return RenderInfo();
}

bool AVSWrapper::RenderInfo()
{
	m_NextFrame = 0;
	if (! m_Clip) {
		m_Info = Info();
		return false;
	}
	const VideoInfo & videoInfo = m_Clip->GetVideoInfo();
	m_Info.hasVideo = videoInfo.HasVideo();
	if (m_Info.hasVideo) {
		m_Info.width      = videoInfo.width;
		m_Info.height     = videoInfo.height;
		m_Info.frameCount = videoInfo.num_frames;
		m_Info.frameRate  = 1.0 * videoInfo.fps_numerator / videoInfo.fps_denominator;
	}
	m_Info.hasAudio = videoInfo.HasAudio();
	if (m_Info.hasAudio) {
		m_Info.noOfChannels = videoInfo.AudioChannels();
	}
	return (m_Info.hasVideo || m_Info.hasAudio);
}

const String & AVSWrapper::GetError() const
{
	return m_Error;
}

const AVSWrapper::Info & AVSWrapper::GetInfo() const
{
	return m_Info;
}

HBITMAP AVSWrapper::GetNextFrame()
{
	if (! m_Clip) return NULL;
	if (m_NextFrame >= m_Clip->GetVideoInfo().num_frames) return NULL;
	
	if (! m_Clip->GetVideoInfo().IsColorSpace(VideoInfo::CS_BGR24)) {
		try {
			AVSValue argValue[1]; argValue[0] = m_Clip;
			AVSValue clip = m_ScriptEnv->Invoke("ConvertToRGB24", AVSValue(argValue, 1));
			if (! clip.IsClip()) return NULL;
			m_Clip = clip.AsClip();
		} catch (AvisynthError & error) {
			m_Error = String().fromAnsi(error.msg).firstLine();
			m_Clip = NULL;
		} catch (...) {
			return NULL;
		}
	}

	try {
		PVideoFrame frame = m_Clip->GetFrame(m_NextFrame, m_ScriptEnv);
		BITMAPINFO bitmapInfo;
		ZeroMemory(& bitmapInfo, sizeof(BITMAPINFO));
		bitmapInfo.bmiHeader.biSize = sizeof(BITMAPINFO);
		bitmapInfo.bmiHeader.biWidth = m_Clip->GetVideoInfo().width;
		bitmapInfo.bmiHeader.biHeight = m_Clip->GetVideoInfo().height; 
		bitmapInfo.bmiHeader.biPlanes = 1;
		bitmapInfo.bmiHeader.biBitCount = 24;
		bitmapInfo.bmiHeader.biCompression = BI_RGB;
		HDC dc = GetDC(NULL);
		HBITMAP bitmap = CreateDIBitmap(dc, & bitmapInfo.bmiHeader, CBM_INIT, frame->GetReadPtr(), & bitmapInfo, 0);
		ReleaseDC(NULL, dc);
		if (bitmap) m_NextFrame++;
		return bitmap;
	} catch (AvisynthError & error) {
		m_Error = String().fromAnsi(error.msg).firstLine();
		return NULL;
	} catch (...) {
		return NULL;
	}
}
	
bool AVSWrapper::IsScript(const String & fileName)
{
	if (fileName.length() >= 4) {
		return ! lstrcmpi(fileName.substr(fileName.length() - 4).c_str(), TEXT(".avs"));
	} else {
		return false;
	}
}

