#include "langcode.h"
#include "project.h"
#include "group.h"

Group::Group(Project * project) : ManagablePtr()
{
	m_Project            = project;
	m_FirstSubtitleOn    = false;
	m_MenuAspectRatio    = ASPECTRATIO_4_3;
	m_HasMenu            = false;
	m_Menu               = new Menu(this, MENU_ROOT);
	m_MenuType           = 0;
	m_CombinedMenu       = new Menu(this, MENU_COMBINED);
	m_HasAudioMenu       = false;
	m_AudioMenu          = new Menu(this, MENU_AUDIO);
	m_HasSubtitleMenu    = false;
	m_SubtitleMenu       = new Menu(this, MENU_SUBTITLE);
	m_Length             = 0.0;
	m_IsTitleSelected    = 0;
	m_SelectedVideoCount = 0;
	m_FirstSelectedVideo = -1;
	m_LastSelectedVideo  = -1;
	m_SelectedAudio      = -1;
	m_SelectedSubtitle   = -1;
	m_View               = 0;
	m_Index              = -1;
}

Group::~Group()
{
}

Project * Group::GetProject() const
{
	return m_Project;
}

bool Group::GetFirstSubtitleOn() const
{
	return m_FirstSubtitleOn;
}

bool Group::SetFirstSubtitleOn(bool firstSubtitleOn, bool setChanged)
{
	if (m_FirstSubtitleOn != firstSubtitleOn) {
		m_FirstSubtitleOn = firstSubtitleOn;
		if (setChanged) SetChanged();
	}
	return true;
}

int Group::GetMenuAspectRatio() const
{
	return m_MenuAspectRatio;
}

String Group::GetMenuAspectRatioStr() const
{
	return (m_MenuAspectRatio == ASPECTRATIO_16_9 ? S("16:9") : S("4:3"));
}

bool Group::SetMenuAspectRatio(int aspectRatio, bool setChanged)
{
	if (0 <= aspectRatio && aspectRatio <= ASPECTRATIO_16_9) {
		if (m_MenuAspectRatio != aspectRatio) {
			m_MenuAspectRatio = aspectRatio;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

bool Group::CanAddVideo() const
{
	return true;
}

bool Group::AddVideo(VideoPtr & video, bool setChanged)
{
	if (CanAddVideo()) {
		m_Videos.push_back(video);
		if (setChanged) {
			video->AddTracks(true);
			video->SetFileSelected(true);
			video->SetTrackSelected(true);
			video->SetTitleSelected(true);
			video->SetChanged();
		}
		return true;
	}
	return false;
}

bool Group::IsVideoIndexValid(int index) const
{
	return (0 <= index && index < (int) m_Videos.size());
}

int Group::GetVideoCount() const
{
	return (int) m_Videos.size();
}

VideoPtr Group::GetVideo(int index) const
{
	if (IsVideoIndexValid(index)) {
		return m_Videos[index];
	} else {
		return VideoPtr(0);
	}
}

bool Group::CanMoveSelectedVideosUp() const
{
	return (m_SelectedVideoCount && m_FirstSelectedVideo > 0);
}

bool Group::MoveSelectedVideosUp(bool setChanged)
{
	if (CanMoveSelectedVideosUp()) {
		for (int index = 1; index < (int) m_Videos.size(); index++) {
			if (m_Videos[index]->IsSelected()) {
				VideoPtr video = m_Videos[index];
				m_Videos[index] = m_Videos[index - 1];
				m_Videos[index - 1] = video;
				if (m_Videos[index]->GetPartNo() != video->GetPartNo()) {
					bool merged = video->IsMergedWithNext();
					video->SetMergeWithNext(m_Videos[index]->IsMergedWithNext());
					m_Videos[index]->SetMergeWithNext(merged, false);
				}
			}
		}
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Group::CanMoveSelectedVideosDown() const
{
	return (m_SelectedVideoCount && m_LastSelectedVideo < (int) m_Videos.size() - 1);
}

bool Group::MoveSelectedVideosDown(bool setChanged)
{
	if (CanMoveSelectedVideosDown()) {
		for (int index = (int) m_Videos.size() - 2; index >= 0 ; index--) {
			if (m_Videos[index]->IsSelected()) {
				VideoPtr video = m_Videos[index];
				m_Videos[index] = m_Videos[index + 1];
				m_Videos[index + 1] = video;
				if (m_Videos[index]->GetPartNo() != m_Videos[index + 1]->GetPartNo()) {
					bool merged = video->IsMergedWithNext();
					video->SetMergeWithNext(m_Videos[index]->IsMergedWithNext());
					m_Videos[index]->SetMergeWithNext(merged, false);
				}
			}
		}
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Group::CanMergeSelectedVideos() const
{
	for (int index = (int) m_Videos.size() - 2; index >= 0 ; index--) {
		if (m_Videos[index]->IsSelected() && m_Videos[index + 1]->IsSelected() &&
			! m_Videos[index]->IsMergedWithNext()) return true;
	}
	return false;
}

bool Group::MergeSelectedVideos(bool setChanged)
{
	bool merged = false;
	for (int index = (int) m_Videos.size() - 2; index >= 0 ; index--) {
		if (m_Videos[index]->IsSelected() && m_Videos[index + 1]->IsSelected() &&
			! m_Videos[index]->IsMergedWithNext()) {
			m_Videos[index]->SetMergeWithNext(true, false);
			merged = true;
		}
	}
	if (merged && setChanged) SetChanged();
	return merged;
}

bool Group::CanSplitSelectedVideos() const
{
	for (int index = (int) m_Videos.size() - 2; index >= 0 ; index--) {
		if (m_Videos[index]->IsSelected() && m_Videos[index + 1]->IsSelected() &&
			m_Videos[index]->IsMergedWithNext()) return true;
	}
	return false;
}

bool Group::SplitSelectedVideos(bool setChanged)
{
	bool split = false;
	for (int index = (int) m_Videos.size() - 2; index >= 0 ; index--) {
		if (m_Videos[index]->IsSelected() && m_Videos[index + 1]->IsSelected() &&
			m_Videos[index]->IsMergedWithNext()) {
			m_Videos[index]->SetMergeWithNext(false, false);
			split = true;
		}
	}
	if (split && setChanged) SetChanged();
	return split;
}

bool Group::CanDeleteSelectedVideos() const
{
	return (m_SelectedVideoCount > 0);
}

bool Group::DeleteSelectedVideos(bool setChanged)
{	
	bool deleted = false;
	for (VideoVector::iterator videoIterator = m_Videos.begin(); videoIterator != m_Videos.end();) {
		if ((*videoIterator)->IsSelected()) {
			videoIterator = m_Videos.erase(videoIterator);
			deleted = true;
		} else {
			videoIterator++;
		}
	}
	if (deleted && setChanged) SetChanged();
	return deleted;
}

bool Group::CanAddAudio() const
{
	return (m_AudioLanguages.size() < 8); 
}

bool Group::AddAudio(int language, bool setChanged)
{
	if (CanAddAudio()) {
		if (0 <= language && language < LANGUAGE_COUNT) {
			bool first = ! m_AudioLanguages.size();
			m_AudioLanguages.push_back(language);
			m_AudioSelected.push_back(setChanged);
			for (VideoVector::iterator videoIterator = m_Videos.begin(); 
				videoIterator != m_Videos.end(); videoIterator++) {
				(*videoIterator)->AddAudio(setChanged);
			}
			if (setChanged) {
				for (int i = 0; i < (int) m_AudioSelected.size() - 1; i++) 
					m_AudioSelected[i] = false;
				SetSelectedAudio(GetAudioCount() - 1);
				SetChanged();
			}
			return true;
		}
	}
	return false;
}

int Group::GetAudioCount() const
{
	return (int) m_AudioLanguages.size();
}

bool Group::IsAudioIndexValid(int index) const
{
	return (0 <= index && index < (int) m_AudioLanguages.size());
}

bool Group::SetAudioLanguage(int index, int language, bool setChanged)
{
	if (IsAudioIndexValid(index)) {
		if (0 <= language && language < LANGUAGE_COUNT) {
			if (m_AudioLanguages[index] != language) {
				m_AudioLanguages[index] = language;
				if (setChanged) SetChanged();
			}
			return true;
		}
	}
	return false;
}

int Group::GetAudioLanguage(int index) const
{
	if (IsAudioIndexValid(index)) {
		return m_AudioLanguages[index];
	}
	return 0;
}

bool Group::CanMoveAudioUp(int index) const
{
	return (0 < index && index < (int) m_AudioLanguages.size());
}

bool Group::MoveAudioUp(int index, bool setChanged)
{
	if (CanMoveAudioUp(index)) {
		int language = m_AudioLanguages[index];
		bool selected = m_AudioSelected[index];
		m_AudioLanguages[index] = m_AudioLanguages[index - 1];
		m_AudioSelected[index] = m_AudioSelected[index - 1];
		m_AudioLanguages[index - 1] = language;
		m_AudioSelected[index - 1] = selected;
		for (VideoVector::iterator videoIterator = m_Videos.begin(); 
			videoIterator != m_Videos.end(); videoIterator++) {
			(*videoIterator)->MoveAudioUp(index);
		}
		if (index == m_SelectedAudio) m_SelectedAudio--;
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Group::CanMoveAudioDown(int index) const
{
	return (0 <= index && index < (int) m_AudioLanguages.size() - 1);
}

bool Group::MoveAudioDown(int index, bool setChanged)
{
	if (CanMoveAudioDown(index)) {
		int language = m_AudioLanguages[index];
		bool selected = m_AudioSelected[index];
		m_AudioLanguages[index] = m_AudioLanguages[index + 1];
		m_AudioSelected[index] = m_AudioSelected[index + 1];
		m_AudioLanguages[index + 1] = language;
		m_AudioSelected[index + 1] = selected;
		for (VideoVector::iterator videoIterator = m_Videos.begin(); 
			videoIterator != m_Videos.end(); videoIterator++) {
			(*videoIterator)->MoveAudioDown(index);
		}
		if (index == m_SelectedAudio) m_SelectedAudio++;
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Group::CanDeleteAudio(int index) const
{
	return IsAudioIndexValid(index);
}

bool Group::DeleteAudio(int index, bool setChanged)
{
	if (IsAudioIndexValid(index)) {
		m_AudioLanguages.erase(m_AudioLanguages.begin() + index);
		m_AudioSelected.erase(m_AudioSelected.begin() + index);
		for (VideoVector::iterator videoIterator = m_Videos.begin(); 
			videoIterator != m_Videos.end(); videoIterator++) {
			(*videoIterator)->DeleteAudio(index);
		}
		if (setChanged) {
			SetSelectedAudio(-1);
			SetChanged();
		}
		return true;
	}
	return false;
}

bool Group::CanAddSubtitle() const
{
	return (m_SubtitleLanguages.size() < 32);
}

int Group::GetSelectedAudio() const
{
	return m_SelectedAudio;
}

bool Group::SetSelectedAudio(int selectedAudio)
{
	if (selectedAudio < GetAudioCount()) {
		if (selectedAudio < 0) {
			m_SelectedAudio = -1;
		} else {
			m_SelectedAudio = selectedAudio;
		}
		return true;
	}
	return false;
}

bool Group::IsAudioSelected() const
{
	return (m_SelectedAudio >= 0);
}

bool Group::AddSubtitle(int language, bool setChanged)
{
	if (CanAddSubtitle()) {
		if (0 <= language && language < LANGUAGE_COUNT) {
			m_SubtitleLanguages.push_back(language);
			m_SubtitleSelected.push_back(setChanged);
			for (VideoVector::iterator videoIterator = m_Videos.begin(); 
				videoIterator != m_Videos.end(); videoIterator++) {
				(*videoIterator)->AddSubtitle(setChanged);
			}
			if (setChanged) {
				for (int i = 0; i < (int) m_SubtitleSelected.size() - 1; i++) 
					m_SubtitleSelected[i] = false;
				SetSelectedSubtitle(GetSubtitleCount() - 1);
				SetChanged();
			}
			return true;
		}
	}
	return false;
}

int Group::GetSubtitleCount() const
{
	return (int) m_SubtitleLanguages.size();
}

bool Group::IsSubtitleIndexValid(int index) const
{
	return (0 <= index && index < (int) m_SubtitleLanguages.size());
}

bool Group::SetSubtitleLanguage(int index, int language, bool setChanged)
{
	if (IsSubtitleIndexValid(index)) {
		if (0 <= language && language < LANGUAGE_COUNT) {
			if (m_SubtitleLanguages[index] != language) {
				m_SubtitleLanguages[index] = language;
				if (setChanged) SetChanged();
			}
			return true;
		}
	}
	return false;
}

int Group::GetSubtitleLanguage(int index) const
{
	if (IsSubtitleIndexValid(index)) {
		return m_SubtitleLanguages[index];
	}
	return 0;
}

bool Group::CanMoveSubtitleUp(int index) const
{
	return (0 < index && index < (int) m_SubtitleLanguages.size());
}

bool Group::MoveSubtitleUp(int index, bool setChanged)
{
	if (CanMoveSubtitleUp(index)) {
		int language = m_SubtitleLanguages[index];
		bool selected = m_SubtitleSelected[index];
		m_SubtitleLanguages[index] = m_SubtitleLanguages[index - 1];
		m_SubtitleSelected[index] = m_SubtitleSelected[index - 1];
		m_SubtitleLanguages[index - 1] = language;
		m_SubtitleSelected[index - 1] = selected;
		for (VideoVector::iterator videoIterator = m_Videos.begin(); 
			videoIterator != m_Videos.end(); videoIterator++) {
			(*videoIterator)->MoveSubtitleUp(index);
		}
		if (index == m_SelectedSubtitle) m_SelectedSubtitle--;
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Group::CanMoveSubtitleDown(int index) const
{
	return (0 <= index && index < (int) m_SubtitleLanguages.size() - 1);
}

bool Group::MoveSubtitleDown(int index, bool setChanged)
{
	if (CanMoveSubtitleDown(index)) {
		int language = m_SubtitleLanguages[index];
		bool selected = m_SubtitleSelected[index];
		m_SubtitleLanguages[index] = m_SubtitleLanguages[index + 1];
		m_SubtitleSelected[index] = m_SubtitleSelected[index + 1];
		m_SubtitleLanguages[index + 1] = language;
		m_SubtitleSelected[index + 1] = selected;
		for (VideoVector::iterator videoIterator = m_Videos.begin(); 
			videoIterator != m_Videos.end(); videoIterator++) {
			(*videoIterator)->MoveSubtitleDown(index);
		}
		if (index == m_SelectedSubtitle) m_SelectedSubtitle++;
		if (setChanged) SetChanged();
		return true;
	}
	return false;
}

bool Group::CanDeleteSubtitle(int index) const
{
	return IsSubtitleIndexValid(index);
}

bool Group::DeleteSubtitle(int index, bool setChanged)
{
	if (IsSubtitleIndexValid(index)) {
		m_SubtitleLanguages.erase(m_SubtitleLanguages.begin() + index);
		m_SubtitleSelected.erase(m_SubtitleSelected.begin() + index);
		for (VideoVector::iterator videoIterator = m_Videos.begin(); 
			videoIterator != m_Videos.end(); videoIterator++) {
			(*videoIterator)->DeleteSubtitle(index);
		}
		if (setChanged) {
			SetSelectedSubtitle(-1);
			SetChanged();
		}
		return true;
	}
	return false;
}

int Group::GetSelectedSubtitle() const
{
	return m_SelectedSubtitle;
}

bool Group::SetSelectedSubtitle(int selectedSubtitle)
{
	if (selectedSubtitle < GetSubtitleCount()) {
		if (selectedSubtitle < 0) {
			m_SelectedSubtitle = -1;
		} else {
			m_SelectedSubtitle = selectedSubtitle;
		}
		return true;
	}
	return false;
}

bool Group::IsSubtitleSelected() const
{
	return (m_SelectedSubtitle >= 0);
}

bool Group::HasMenu() const
{
	return m_HasMenu;
}

bool Group::SetMenu(bool hasMenu, bool setChanged)
{
	if (m_HasMenu != hasMenu) {
		m_HasMenu = hasMenu;
		if (setChanged) SetChanged();
	}
	return true;
}

MenuPtr Group::GetMenu() const
{
	return m_Menu;
}

int Group::GetMenuType() const
{
	return m_MenuType;
}

bool Group::SetMenuType(int menuType, bool setChanged)
{
	if (0 <= menuType && menuType < MENUTYPE_MAX) {
		if (m_MenuType != menuType) {
			m_MenuType = menuType;
			if (setChanged) SetChanged();
		}
		return true;
	}
	return false;
}

MenuPtr Group::GetCombinedMenu() const
{
	return m_CombinedMenu;
}

bool Group::HasAudioMenu() const
{
	return m_HasAudioMenu;
}

bool Group::SetAudioMenu(bool hasAudioMenu, bool setChanged)
{
	if (m_HasAudioMenu != hasAudioMenu) {
		m_HasAudioMenu = hasAudioMenu;
		if (setChanged) SetChanged();
	}
	return true;
}

MenuPtr Group::GetAudioMenu() const
{
	return m_AudioMenu;
}

bool Group::HasSubtitleMenu() const
{
	return m_HasSubtitleMenu;
}

bool Group::SetSubtitleMenu(bool hasSubtitleMenu, bool setChanged)
{
	if (m_HasSubtitleMenu != hasSubtitleMenu) {
		m_HasSubtitleMenu = hasSubtitleMenu;
		if (setChanged) SetChanged();
	}
	return true;
}

MenuPtr Group::GetSubtitleMenu() const
{
	return m_SubtitleMenu;
}

void Group::SetChanged()
{
	m_Project->SetChanged();
}


int Group::GetIndex() const
{
	return m_Index;
}

void Group::SetIndex(int index)
{
	m_Index = index;
}

String Group::GetIdent()
{
	if (m_Index < 0) {
		return S("");
	} else {
		return S("Group ") + String(GetIndex() + 1) + S(": ");
	}
}

String Group::Recalculate()
{
	String error;

	TitlePtr title;
	m_Titles.clear();
	int titleNo = 0;
	int partNo  = 1;
	bool lastIsMerged = false;
	for (int vindex = 0; vindex < GetVideoCount(); vindex++) {
		Video * video = GetVideo(vindex);
		if (lastIsMerged) {
			partNo++;
		} else {
			partNo = 1;
			titleNo++;
			title = new Title(this);
			m_Titles.push_back(title);
		}
		title->AddVideo(video);
		video->SetTitle(title);
		video->SetTitleNo(titleNo);
		video->SetPartNo(partNo);
		video->SetMergeWithPrev(lastIsMerged);
		lastIsMerged = video->IsMergedWithNext();
	}

	bool cannotSelectTitle = false;
	m_IsTitleSelected = false;
	m_SelectedVideoCount = 0;
	m_FirstSelectedVideo = -1;
	m_LastSelectedVideo  = -1;
	int index = 0;
	m_Length = 0.0;
	m_SubStreamRate = 0;
	for (int tindex = 0; tindex < GetTitleCount(); tindex++) {
		Title * title = GetTitle(tindex);
		title->SetIndex(tindex);
		error += title->Recalculate();
		if (title->IsSelected()) {
			m_IsTitleSelected = true;
		}
		for (int vindex = 0; vindex < title->GetVideoCount(); vindex++) {
			Video * video = title->GetVideo(vindex);
			if (video->IsSelected()) {
				m_SelectedVideoCount++;
				if (m_FirstSelectedVideo < 0) {
					m_FirstSelectedVideo = index;
				}
				m_LastSelectedVideo = index;
				if (! title->IsSelected()) {
					cannotSelectTitle = true;
				}
			} 
			m_Length += video->GetLength();
			if (m_SubStreamRate < video->GetSubStreamRate()) {
				m_SubStreamRate = video->GetSubStreamRate();
			}
			index++;
		}
	}
	if (cannotSelectTitle) {
		m_IsTitleSelected = false;
	}

	if (HasMenu()) error += m_Menu->Recalculate();
	switch (GetMenuType()) {
	case MENUTYPE_COMBINED:
		error += m_CombinedMenu->Recalculate();
		break;
	case MENUTYPE_SEPARATE:
		if (HasAudioMenu()) error += m_AudioMenu->Recalculate();
		if (HasSubtitleMenu()) error += m_SubtitleMenu->Recalculate();
		break;
	}

	if (! GetTitleCount() && ! HasMenu()) {
		error += (GetIdent() + S("Neither titles nor menus are defined.\n"));
	}
	return error;
}

int Group::GetTitleCount() const
{
	return (int) m_Titles.size();
}

bool Group::IsTitleIndexValid(int index) const
{
	return (0 <= index && index <= (int) m_Titles.size());
}

TitlePtr Group::GetTitle(int index) const
{
	if (IsTitleIndexValid(index)) {
		return m_Titles[index];
	} else {
		return TitlePtr();
	}
}

int Group::GetView() const
{
	return m_View;
}

void Group::SetView(int view)
{
	if (0 <= view && view < VIEW_MAX) {
		m_View = view;
	}
}

bool Group::IsVideoTrackSelected() const
{
	return m_VideoTrackSelected;
}

void Group::SetVideoTrackSelected(bool selected)
{
	m_VideoTrackSelected = selected;
}

bool Group::IsAudioTrackSelected(int index) const
{
	return m_AudioSelected[index];
}

void Group::SetAudioTrackSelected(int index, bool selected)
{
	m_AudioSelected[index] = selected;
}

bool Group::IsSubtitleTrackSelected(int index) const
{
	return m_SubtitleSelected[index];
}

void Group::SetSubtitleTrackSelected(int index, bool selected)
{
	m_SubtitleSelected[index] = selected;
}

double Group::GetLength() const
{
	return m_Length;
}

int Group::GetSubStreamRate() const
{
	return m_SubStreamRate;
}

void Group::GetDescription(StringVectorVector & strings) const
{
	StringVector line;
	line.push_back(String());
	line.push_back(String(GetTitleCount()) + S(" titles (") + String((int) m_Videos.size()) + S(" videos)"));
	line.push_back(S(" - length: "));
	line.push_back(String().fromLength(m_Length));
	strings.push_back(line);

	line.clear();
	line.push_back(S("Audio tracks: "));
	String languages;
	bool first = true;
	for (std::vector<int>::const_iterator audioLanguageIterator = m_AudioLanguages.begin();
		audioLanguageIterator != m_AudioLanguages.end(); audioLanguageIterator++) {
		if (first) {
			first = false;
		} else {
			languages += String(TEXT(", "));
		}
		languages += String(LanguageName[*audioLanguageIterator]);
	}
	if (first) {
		languages = S("-");
	}
	line.push_back(languages);
	strings.push_back(line);

	line.clear();
	line.push_back(S("Subtitles: "));
	languages.clear();
	first = true;
	for (std::vector<int>::const_iterator subtitleLanguageIterator = m_SubtitleLanguages.begin();
		subtitleLanguageIterator != m_SubtitleLanguages.end(); subtitleLanguageIterator++) {
		if (first) {
			first = false;
		} else {
			languages += String(TEXT(", "));
		}
		languages += String(LanguageName[*subtitleLanguageIterator]);
	}
	if (first) {
		languages = S("-");
	}
	line.push_back(languages);
	strings.push_back(line);

	line.clear();
	line.push_back(S("Menus: "));
	if (m_HasMenu) {
		String s = S("Root");
		if (m_MenuType == MENUTYPE_COMBINED) {
			s += S(", Combined");
		} else if (m_MenuType == MENUTYPE_SEPARATE) {
			if (m_HasAudioMenu) {
				s += S(", Audio");
			}
			if (m_HasSubtitleMenu) {
				s += S(", Subtitle");
			}
		}
		line.push_back(s);
	} else {
		line.push_back(S("-"));
	}
	strings.push_back(line);
}

String Group::GetShortDescr(int index) const
{
	String result = S("Group ") + String(index + 1) + S(": ") + 
		String(GetTitleCount()) + S(" titles (") + String((int) m_Videos.size()) + S(" videos)\n") +
		S("Sum length: ") + String().fromLength(m_Length) + S("\n") +
		String((int) m_AudioLanguages.size()) + S(" audio tracks, ") + 
		String((int) m_SubtitleLanguages.size()) + S(" subtitles");
	return result;
}

bool Group::Load(csv_file_t csv, int & lineno, int nest)
{
	bool ok = true;

	int field = 1;
	ok = ok && SetFirstSubtitleOn(csv_get_int(csv, lineno, ++field, m_FirstSubtitleOn ? 1 : 0) != 0, false);
	ok = ok && SetMenu(csv_get_int(csv, lineno, ++field, m_HasMenu ? 1 : 0) != 0, false);
	ok = ok && SetMenuType(csv_get_int(csv, lineno, ++field, m_MenuType), false);
	ok = ok && SetAudioMenu(csv_get_int(csv, lineno, ++field, m_HasAudioMenu ? 1 : 0) != 0, false);
	ok = ok && SetSubtitleMenu(csv_get_int(csv, lineno, ++field, m_HasSubtitleMenu ? 1 : 0) != 0, false);
	int selectedAudio = csv_get_int(csv, lineno, ++field, 0);
	int selectedSubtitle = csv_get_int(csv, lineno, ++field, 0);
	SetView(csv_get_int(csv, lineno, ++field, m_View));
	SetVideoTrackSelected(csv_get_int(csv, lineno, ++field, m_View) != 0);
	ok = ok && SetMenuAspectRatio(csv_get_int(csv, lineno, ++field, m_MenuAspectRatio), false);
	
	bool gotVideos = false;
	lineno++;
	while (ok && lineno < csv->count && csv_get_int(csv, lineno, 0, 0) == nest) {
		const char * type = csv_get_str(csv, lineno, 1, "");

		if (! strcmp(type, "audio")) {
			ok = ok && ! gotVideos;
			int audioLanguage = csv_get_int(csv, lineno, 2, 0);
			ok = ok && (audioLanguage < LANGUAGE_COUNT);
			m_AudioLanguages.push_back(audioLanguage);
			int selected = csv_get_int(csv, lineno, 3, 0);
			m_AudioSelected.push_back(selected != 0);
			lineno++;
	
		} else if (! strcmp(type, "subtitle")) {
			ok = ok && ! gotVideos;
			int subtitleLanguage = csv_get_int(csv, lineno, 2, 0);
			ok = ok && (subtitleLanguage < LANGUAGE_COUNT);
			m_SubtitleLanguages.push_back(subtitleLanguage);
			int selected = csv_get_int(csv, lineno, 3, 0);
			m_SubtitleSelected.push_back(selected != 0);
			lineno++;

		} else if (! strcmp(type, "video")) {
			gotVideos = true;
			VideoPtr video = new Video(this);
			if (video->Load(csv, lineno, nest + 1)) {
				ok = ok && video->GetAudioCount() == m_AudioLanguages.size();
				ok = ok && video->GetSubtitleCount() == m_SubtitleLanguages.size();
				ok = ok && AddVideo(video, false);
			} else {
				ok = false;
			}

		} else if (! strcmp(type, "rootmenu")) {
			ok = ok && m_Menu->Load(csv, lineno, nest + 1);

		} else if (! strcmp(type, "combinedmenu")) {
			ok = ok && m_CombinedMenu->Load(csv, lineno, nest + 1);

		} else if (! strcmp(type, "audiomenu")) {
			ok = ok && m_AudioMenu->Load(csv, lineno, nest + 1);

		} else if (! strcmp(type, "subtitlemenu")) {
			ok = ok && m_SubtitleMenu->Load(csv, lineno, nest + 1);

		} else {
			lineno++;
		}
	}

	ok = ok && SetSelectedAudio(selectedAudio);
	ok = ok && SetSelectedSubtitle(selectedSubtitle);

	return ok;
}
void Group::Save(FILE * f, int nest) const
{
	fprintf(f, "%d,group,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n", nest, 
		(m_FirstSubtitleOn ? 1 : 0),
		(m_HasMenu ? 1 : 0),
		m_MenuType,
		(m_HasAudioMenu ? 1 : 0),
		(m_HasSubtitleMenu ? 1 : 0),
		m_SelectedAudio,
		m_SelectedSubtitle,
		m_View,
		m_VideoTrackSelected,
		m_MenuAspectRatio);

	for (int i = 0; i < (int) m_AudioLanguages.size(); i++) {
		fprintf(f, "%d,audio,%d,%d\n", nest + 1, m_AudioLanguages[i], (m_AudioSelected[i] ? 1 : 0));
	}

	for (int i = 0; i < (int) m_SubtitleLanguages.size(); i++) {
		fprintf(f, "%d,subtitle,%d,%d\n", nest + 1, m_SubtitleLanguages[i], (m_SubtitleSelected[i] ? 1 : 0));
	}

	for (VideoVector::const_iterator videoIterator = m_Videos.begin(); 
		videoIterator != m_Videos.end(); videoIterator++) {
		(*videoIterator)->Save(f, nest + 1);
	}

	if (m_HasMenu) {
		m_Menu->Save(f, nest + 1);
		switch (m_MenuType) {
		case MENUTYPE_COMBINED:
			m_CombinedMenu->Save(f, nest + 1);
			break;
		case MENUTYPE_SEPARATE:
			if (m_HasAudioMenu) m_AudioMenu->Save(f, nest + 1);
			if (m_HasSubtitleMenu) m_SubtitleMenu->Save(f, nest + 1);
			break;
		}
	}
}
