#include "menudialog.h"
#include "menuitem.h"
#include "menudrawer.h"
#include "main.h"
#include "resource.h"
#include "drawitem.h"
#include "addmenuitemdialog.h"
#include "tools.h"
#include "avswrapper.h"
#include "aspectpreview.h"

#include "winutils/winutils.h"

#include <stdio.h>

MenuDialog::MenuDialog(MainDialog * mainDialog, MenuPtr menu, int groupIndex)
{
	m_MainDialog = mainDialog;
	m_Menu       = menu;
	m_GroupIndex = groupIndex;
	m_Bitmap     = NULL;
	m_DragOp     = DRAG_SELECT;
	m_IsDragging = false;
}

MenuDialog::~MenuDialog()
{
	if (m_Bitmap) DeleteObject(m_Bitmap);
}

int MenuDialog::DoModal()
{
	return Dialog::DoModal(hInstance, MAKEINTRESOURCE(IDD_MENU), m_MainDialog->hWnd());
}

void MenuDialog::UpdateBitmap()
{
	if (m_Bitmap) DeleteObject(m_Bitmap);
	String imgFileName = m_Menu->GetBackgroundImage();
	if (imgFileName.length()) {
		HCURSOR origCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
		int width  = m_Menu->GetGroup()->GetProject()->GetVideoWidth();
		int height = m_Menu->GetGroup()->GetProject()->GetVideoHeight();
		AVSWrapper avs;
		avs.ImageSource(imgFileName, width, height);
		m_Bitmap = avs.GetNextFrame();
		if (m_Bitmap) {
			HDC dc = GetDC(hWnd());
			HDC imageDC = CreateCompatibleDC(dc);
			SelectObject(imageDC, m_Bitmap);
			double red = 0.0, green = 0.0, blue = 0.0;
			double count = 0;
			for (int x = 0; x < width; x++) {
				for (int y = 0; y < height; y++) {
					COLORREF pixel = GetPixel(imageDC, x, y);
					red   += GetRValue(pixel);
					green += GetGValue(pixel);
					blue  += GetBValue(pixel);
					count++;
				}
			}
			COLORREF bgColor = RGB((int) (red / count), (int) (green / count), (int) (blue / count));
			m_Menu->SetBackgroundColor(bgColor);
			DeleteDC(imageDC);
			ReleaseDC(hWnd(), dc);
		}			
		SetCursor(origCursor);
	}
}

void MenuDialog::Update(bool reset)
{
	String dialogTitle = S("Edit ");
	switch (m_Menu->GetType()) {
	case MENU_ROOT:
		dialogTitle += S("root");
		break;
	case MENU_COMBINED:
		dialogTitle += S("audio + subtitle");
		break;
	case MENU_AUDIO:
		dialogTitle += S("audio");
		break;
	case MENU_SUBTITLE:
		dialogTitle += S("subtitle");
		break;
	}
	dialogTitle += S(" menu of group ") + String(m_GroupIndex + 1);
	SetWindowText(hWnd(), dialogTitle.c_str());

	InvalidateRect(GetDlgItem(hWnd(), IDC_PREVIEW), NULL, FALSE);

	SetDlgItemText(hWnd(), IDC_BGIMAGEFILE, m_Menu->GetBackgroundImage().c_str());
	SendDlgItemMessage(hWnd(), IDC_ASPECT_4_3, BM_SETCHECK, 
		(m_Menu->GetGroup()->GetMenuAspectRatio() == ASPECTRATIO_4_3 ? BST_CHECKED : BST_UNCHECKED), 0);
	SendDlgItemMessage(hWnd(), IDC_ASPECT_16_9, BM_SETCHECK, 
		(m_Menu->GetGroup()->GetMenuAspectRatio() == ASPECTRATIO_16_9 ? BST_CHECKED : BST_UNCHECKED), 0);
	SetWindowLong(GetDlgItem(hWnd(), IDC_VIEWBGCOLOR), GWL_USERDATA, m_Menu->GetBackgroundColor());
	InvalidateRect(GetDlgItem(hWnd(), IDC_VIEWBGCOLOR), NULL, FALSE);
	SetWindowLong(GetDlgItem(hWnd(), IDC_VIEWNORMALCOLOR), GWL_USERDATA, m_Menu->GetNormalColor());
	InvalidateRect(GetDlgItem(hWnd(), IDC_VIEWNORMALCOLOR), NULL, FALSE);
	SetWindowLong(GetDlgItem(hWnd(), IDC_VIEWHILITECOLOR), GWL_USERDATA, m_Menu->GetHiliteColor());
	InvalidateRect(GetDlgItem(hWnd(), IDC_VIEWHILITECOLOR), NULL, FALSE);
	SetWindowLong(GetDlgItem(hWnd(), IDC_VIEWSELECTCOLOR), GWL_USERDATA, m_Menu->GetSelectColor());
	InvalidateRect(GetDlgItem(hWnd(), IDC_VIEWSELECTCOLOR), NULL, FALSE);

	bool redraw = false;
	SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, WM_SETREDRAW, FALSE, 0);
	if (reset) {
		SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_RESETCONTENT, 0, 0);
		redraw = true;
	}
	int listCount   = (int) SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_GETCOUNT, 0, 0);
	int topIndex    = (int) SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_GETTOPINDEX, 0, 0);
	int caretIndex  = (int) SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_GETCARETINDEX, 0, 0);
	for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) {
		if (index < listCount) {
			SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_SETITEMDATA, (WPARAM) index, (LPARAM) index);
			MEASUREITEMSTRUCT mis;
			mis.CtlID = IDC_MENUITEMLIST;
			mis.itemData = (UINT) index;
			MeasureItem(& mis);
			UINT origHeight = (UINT) SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_GETITEMHEIGHT, index, 0);
			if (mis.itemHeight != origHeight) {
				if (mis.itemHeight < origHeight) redraw = true;
				SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_SETITEMHEIGHT, index, MAKELPARAM(mis.itemHeight, 0));
			}
		} else {
			SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_ADDSTRING, 0, (LPARAM) index);
		}
		if ((SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_GETSEL, index, 0) != 0) != m_Menu->IsMenuItemSelected(index)) {
			SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_SETSEL, m_Menu->IsMenuItemSelected(index), index);
		}
	}
	for (int index = listCount - 1; index >= m_Menu->GetMenuItemCount(); index--) {
		SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_DELETESTRING, index, 0);
		if (topIndex >= index) topIndex = index - 1;
		if (caretIndex >= index) caretIndex = index - 1;
		redraw = true;
	}
	SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_SETTOPINDEX, topIndex, 0);
	SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_SETCARETINDEX, caretIndex, 0);
	SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, WM_SETREDRAW, TRUE, 0);
	InvalidateRect(GetDlgItem(hWnd(), IDC_MENUITEMLIST), NULL, redraw);

	int x          = -1;
	int y          = -1;
	int hasText    = BST_UNCHECKED;
	int normalText = BST_UNCHECKED;
	int hiliteText = BST_UNCHECKED;
	String caption;
	String fontFace;
	int fontSize   = -1;
	int bold       = BST_UNCHECKED;
	int italic     = BST_UNCHECKED;
	int normalRect = BST_UNCHECKED;
	int hiliteRect = BST_UNCHECKED;
	int right      = -1;
	int bottom     = -1;
	int thickness  = -1;
	int roundEdge  = BST_UNCHECKED;
	int hRadius    = -1;
	int vRadius    = -1;
	int fillNormal = BST_UNCHECKED;
	int fillHilite = BST_UNCHECKED;
	int autoResize = BST_UNCHECKED;
	int hMargin    = -1;
	int vMargin    = -1;
	int alignment  = -1;
	bool disableBottomRightAlignment = false;
	bool disableFillNormal = false;
	bool disableFillHilite = false;
	int count = 0;
	for (int i = 0; i < m_Menu->GetMenuItemCount(); i++) {
		MenuItem * menuItem = m_Menu->GetMenuItem(i);
		if (menuItem->IsSelected()) {
			if (! count++) {
				x          = menuItem->GetX();
				y          = menuItem->GetY();
				normalText = (menuItem->GetNormalText() ? BST_CHECKED : BST_UNCHECKED);
				hiliteText = (menuItem->GetHiliteText() ? BST_CHECKED : BST_UNCHECKED);
				caption    = menuItem->GetCaption();
				fontFace   = menuItem->GetFontFace();
				fontSize   = menuItem->GetFontSize();
				bold       = (menuItem->GetBold() ? BST_CHECKED : BST_UNCHECKED);
				italic     = (menuItem->GetItalic() ? BST_CHECKED : BST_UNCHECKED);
				normalRect = (menuItem->GetNormalRect() ? BST_CHECKED : BST_UNCHECKED);
				hiliteRect = (menuItem->GetHiliteRect() ? BST_CHECKED : BST_UNCHECKED);
				right      = menuItem->GetRight();
				bottom     = menuItem->GetBottom();
				thickness  = menuItem->GetThickness();
				hRadius    = menuItem->GetHRadius();
				vRadius    = menuItem->GetVRadius();
				fillNormal = (menuItem->GetFillNormal() ? BST_CHECKED : BST_UNCHECKED);
				fillHilite = (menuItem->GetFillHilite() ? BST_CHECKED : BST_UNCHECKED);
				autoResize = (menuItem->GetAutoResize() ? BST_CHECKED : BST_UNCHECKED);
				hMargin    = menuItem->GetHMargin();
				vMargin    = menuItem->GetVMargin();
				alignment  = menuItem->GetAlignment();
			} else {
				if (x != menuItem->GetX()) 
					x = -1;
				if (y != menuItem->GetY())
					y = -1;
				if (normalText != (menuItem->GetNormalText() ? BST_CHECKED : BST_UNCHECKED))
					normalText = BST_INDETERMINATE;
				if (hiliteText != (menuItem->GetHiliteText() ? BST_CHECKED : BST_UNCHECKED))
					hiliteText = BST_INDETERMINATE;
				if (caption != menuItem->GetCaption())
					caption = S("");
				if (fontFace != menuItem->GetFontFace())
					fontFace = S("");
				if (fontSize != menuItem->GetFontSize())
					fontSize = -1;
				if (bold != (menuItem->GetBold() ? BST_CHECKED : BST_UNCHECKED))
					bold = BST_INDETERMINATE;
				if (italic != (menuItem->GetItalic() ? BST_CHECKED : BST_UNCHECKED))
					italic = BST_INDETERMINATE;
				if (normalRect != (menuItem->GetNormalRect() ? BST_CHECKED : BST_UNCHECKED))
					normalRect = BST_INDETERMINATE;
				if (hiliteRect != (menuItem->GetHiliteRect() ? BST_CHECKED : BST_UNCHECKED))
					hiliteRect = BST_INDETERMINATE;
				if (right != menuItem->GetRight())
					right = -1;
				if (bottom != menuItem->GetBottom())
					bottom = -1;
				if (thickness != menuItem->GetThickness())
					thickness = -1;
				if (hRadius != menuItem->GetHRadius())
					hRadius = -1;
				if (vRadius != menuItem->GetVRadius())
					vRadius = -1;
				if (fillNormal != (menuItem->GetFillNormal() ? BST_CHECKED : BST_UNCHECKED))
					fillNormal = BST_INDETERMINATE;
				if (fillHilite != (menuItem->GetFillHilite() ? BST_CHECKED : BST_UNCHECKED))
					fillHilite = BST_INDETERMINATE;
				if (autoResize != (menuItem->GetAutoResize() ? BST_CHECKED : BST_UNCHECKED))
					autoResize = BST_INDETERMINATE;
				if (hMargin != menuItem->GetHMargin())
					hMargin = -1;
				if (vMargin != menuItem->GetVMargin())
					vMargin = -1;
				if (alignment != menuItem->GetAlignment())
					alignment = -1;
			}
			if (menuItem->GetAutoResize()) {
				disableBottomRightAlignment = true;
			}
			if (! menuItem->GetNormalRect()) {
				disableFillNormal = true;
			}
			if (! menuItem->GetHiliteRect()) {
				disableFillHilite = true;
			}
		}
	}
	EnableWindow(GetDlgItem(hWnd(), IDC_ENTRYGROUP), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_POSPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_POSXPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_POSX), count);
	SetDlgItemText(hWnd(), IDC_POSX, ((x < 0) ? TEXT("") : String(x).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_POSYPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_POSY), count);
	SetDlgItemText(hWnd(), IDC_POSY, ((y < 0) ? TEXT("") : String(y).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_HASTEXT), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_NORMALTEXT), count);
	SendDlgItemMessage(hWnd(), IDC_NORMALTEXT, BM_SETCHECK, normalText, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_HILITETEXT), count);
	SendDlgItemMessage(hWnd(), IDC_HILITETEXT, BM_SETCHECK, hiliteText, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_CAPTION), count);
	SetDlgItemText(hWnd(), IDC_CAPTION, caption.c_str());
	EnableWindow(GetDlgItem(hWnd(), IDC_FONTPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_FONTFACE), count);
	SetDlgItemText(hWnd(), IDC_FONTFACE, fontFace.c_str());
	EnableWindow(GetDlgItem(hWnd(), IDC_SIZEPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_FONTSIZE), count);
	SetDlgItemText(hWnd(), IDC_FONTSIZE, ((fontSize < 0) ? TEXT("") : String(fontSize).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_SIZESUFFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_BOLD), count);
	SendDlgItemMessage(hWnd(), IDC_BOLD, BM_SETCHECK, bold, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_ITALIC), count);
	SendDlgItemMessage(hWnd(), IDC_ITALIC, BM_SETCHECK, italic, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_RECTANGLE), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_NORMALRECT), count);
	SendDlgItemMessage(hWnd(), IDC_NORMALRECT, BM_SETCHECK, normalRect, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_HILITERECT), count);
	SendDlgItemMessage(hWnd(), IDC_HILITERECT, BM_SETCHECK, hiliteRect, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_CORNERPREFIX), count && ! disableBottomRightAlignment);
	EnableWindow(GetDlgItem(hWnd(), IDC_CORNERXPREFIX), count && ! disableBottomRightAlignment);
	EnableWindow(GetDlgItem(hWnd(), IDC_CORNERX), count && ! disableBottomRightAlignment);
	SetDlgItemText(hWnd(), IDC_CORNERX, ((right < 0) ? TEXT("") : String(right).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_CORNERYPREFIX), count && ! disableBottomRightAlignment);
	EnableWindow(GetDlgItem(hWnd(), IDC_CORNERY), count && ! disableBottomRightAlignment);
	SetDlgItemText(hWnd(), IDC_CORNERY, ((bottom < 0) ? TEXT("") : String(bottom).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_THICKNESSPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_THICKNESS), count);
	SetDlgItemText(hWnd(), IDC_THICKNESS, ((thickness < 0) ? TEXT("") : String(thickness).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_THICKNESSUFFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_RADIUSPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_HRADIUSPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_HRADIUS), count);
	SetDlgItemText(hWnd(), IDC_HRADIUS, ((hRadius < 0) ? TEXT("") : String(hRadius).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_VRADIUSPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_VRADIUS), count);
	SetDlgItemText(hWnd(), IDC_VRADIUS, ((vRadius < 0) ? TEXT("") : String(vRadius).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_FILLNORMAL), count && ! disableFillNormal);
	SendDlgItemMessage(hWnd(), IDC_FILLNORMAL, BM_SETCHECK, fillNormal, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_FILLHILITE), count && ! disableFillHilite);
	SendDlgItemMessage(hWnd(), IDC_FILLHILITE, BM_SETCHECK, fillHilite, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_AUTORESIZE), count);
	SendDlgItemMessage(hWnd(), IDC_AUTORESIZE, BM_SETCHECK, autoResize, 0);
	EnableWindow(GetDlgItem(hWnd(), IDC_MARGINPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_HMARGINPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_HMARGIN), count);
	SetDlgItemText(hWnd(), IDC_HMARGIN, ((hMargin < 0) ? TEXT("") : String(hMargin).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_VMARGINPREFIX), count);
	EnableWindow(GetDlgItem(hWnd(), IDC_VMARGIN), count);
	SetDlgItemText(hWnd(), IDC_VMARGIN, ((vMargin < 0) ? TEXT("") : String(vMargin).c_str()));
	EnableWindow(GetDlgItem(hWnd(), IDC_ALIGNPREFIX), count && ! disableBottomRightAlignment);
	EnableWindow(GetDlgItem(hWnd(), IDC_ALIGNMENT), count && ! disableBottomRightAlignment);
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_SETCURSEL, (disableBottomRightAlignment ? -1 : alignment), 0);
	
	m_MainDialog->Update();
}

bool MenuDialog::DialogProc(UINT message, WPARAM wParam, LPARAM lParam)
{
	static bool hasChanged = false;

	switch (message) {
	case WM_INITDIALOG:
		InitDialog();
		return true;

	case WM_COMMAND:
		switch (LOWORD(wParam)) 
		{
		case IDC_BROWSEBGIMAGE:
			if (HIWORD(wParam) == BN_CLICKED) {
				BrowseBackgroundImage();
				return true;
			}
			return false;

		case IDC_CLEARBGIMAGE:
			if (HIWORD(wParam) == BN_CLICKED) {
				m_Menu->SetBackgroundImage(String());
				UpdateBitmap();
				Update();
				return true;
			}
			return false;

		case IDC_CHOOSEBGCOLOR:
			if (HIWORD(wParam) == BN_CLICKED) {
				ChooseMenuColor(MENUCOLOR_BACKGROUND);
				return true;
			}
			return false;

		case IDC_ASPECT_4_3:
			if (HIWORD(wParam) == BN_CLICKED) {
				m_Menu->GetGroup()->SetMenuAspectRatio(ASPECTRATIO_4_3);
				Update();
				return true;
			}
			return false;

		case IDC_ASPECT_16_9:
			if (HIWORD(wParam) == BN_CLICKED) {
				m_Menu->GetGroup()->SetMenuAspectRatio(ASPECTRATIO_16_9);
				Update();
				return true;
			}
			return false;

		case IDC_PREVIEWASPECT:
			if (HIWORD(wParam) == BN_CLICKED) {
				HDC dc = GetDC(hWnd());
				HBITMAP bitmap = CreateCompatibleBitmap(dc, 
					m_Menu->GetGroup()->GetProject()->GetVideoWidth(),
					m_Menu->GetGroup()->GetProject()->GetVideoHeight());
				HDC compDC = CreateCompatibleDC(dc);
				SelectObject(compDC, bitmap);
				if (m_Bitmap) {
					HDC imageDC = CreateCompatibleDC(dc);
					SelectObject(imageDC, m_Bitmap);
					BitBlt(compDC, 0, 0, m_Menu->GetGroup()->GetProject()->GetVideoWidth(), 
						m_Menu->GetGroup()->GetProject()->GetVideoHeight(), imageDC, 0, 0, SRCCOPY);
					DeleteDC(imageDC);
				} else {
					RECT rect;
					rect.left = rect.top = 0;
					rect.right = m_Menu->GetGroup()->GetProject()->GetVideoWidth();
					rect.bottom = m_Menu->GetGroup()->GetProject()->GetVideoHeight();
					HBRUSH brush = CreateSolidBrush(m_Menu->GetBackgroundColor());
					FillRect(compDC, & m_ImageRect, brush);
					DeleteObject(brush);
				}
				MenuDrawer::DrawMenuItems(m_Menu, compDC, 0, 0);
				DeleteDC(compDC);
				ReleaseDC(hWnd(), dc);
				AspectPreview ap(this, bitmap, 
					m_Menu->GetGroup()->GetProject()->GetVideoWidth(),
					m_Menu->GetGroup()->GetProject()->GetVideoHeight(),
					(m_Menu->GetGroup()->GetMenuAspectRatio() == ASPECTRATIO_4_3 ? 4 : 16),
					(m_Menu->GetGroup()->GetMenuAspectRatio() == ASPECTRATIO_4_3 ? 3 : 9));
				ap.DoModal();
				DeleteObject(bitmap);
				return true;
			}
			return false;

		case IDC_CHOOSENORMALCOLOR:
			if (HIWORD(wParam) == BN_CLICKED) {
				ChooseMenuColor(MENUCOLOR_NORMAL);
				return true;
			}
			return false;

		case IDC_CHOOSESELECTCOLOR:
			if (HIWORD(wParam) == BN_CLICKED) {
				ChooseMenuColor(MENUCOLOR_SELECT);
				return true;
			}
			return false;

		case IDC_CHOOSEHILITECOLOR:
			if (HIWORD(wParam) == BN_CLICKED) {
				ChooseMenuColor(MENUCOLOR_HILITE);
				return true;
			}
			return false;

		case IDC_ADDMENUITEM:
			if (HIWORD(wParam) == BN_CLICKED) {
				AddMenuItemDialog addDialog(this, m_Menu, m_GroupIndex);
				if (addDialog.DoModal()) Update();
				return true;
			}
			return false;

		case IDC_DELETEMENUITEM:
			if (HIWORD(wParam) == BN_CLICKED) {
				m_Menu->DeleteSelectedMenuItems();
				Update();
				return true;
			}
			return false;

		case IDC_MENUITEMLIST:
			if (HIWORD(wParam) == LBN_SELCHANGE) {
				OnSelChange();
				return true;
			}
			return false;

		case IDC_POSX:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int x = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_POSX)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetX(x);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_POSY:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int y = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_POSY)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetY(y);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_NORMALTEXT:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool normalText = (SendDlgItemMessage(hWnd(), IDC_NORMALTEXT, BM_GETCHECK, 0, 0) != BST_CHECKED);
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetNormalText(normalText);
				}
				Update();
				return true;
			}
			return false;

		case IDC_CAPTION:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					String caption = String().fromWindow(GetDlgItem(hWnd(), IDC_CAPTION));
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetCaption(caption);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_FONTFACE:
			switch (HIWORD(wParam)) {
			case CBN_SETFOCUS:
				hasChanged = false;
				return true;
			case CBN_SELCHANGE:
			case CBN_EDITCHANGE:
				hasChanged = true;
				return true;
			case CBN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					String fontFace = String().fromWindow(GetDlgItem(hWnd(), IDC_FONTFACE));
					if (GetFontInfo(fontFace.c_str(), NULL, NULL)) {
						for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
							m_Menu->GetMenuItem(index)->SetFontFace(fontFace);
						}
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_FONTSIZE:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int fontSize = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_FONTSIZE)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetFontSize(fontSize);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_BOLD:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool bold = (SendDlgItemMessage(hWnd(), IDC_BOLD, BM_GETCHECK, 0, 0) != BST_CHECKED);
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetBold(bold);
				}
				Update();
				return true;
			}
			return false;

		case IDC_ITALIC:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool italic = (SendDlgItemMessage(hWnd(), IDC_ITALIC, BM_GETCHECK, 0, 0) != BST_CHECKED);
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetItalic(italic);
				}
				Update();
				return true;
			}
			return false;

		case IDC_NORMALRECT:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool normalRect = (SendDlgItemMessage(hWnd(), IDC_NORMALRECT, BM_GETCHECK, 0, 0) != BST_CHECKED);
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetNormalRect(normalRect);
				}
				Update();
				return true;
			}
			return false;

		case IDC_HILITERECT:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool hiliteRect = (SendDlgItemMessage(hWnd(), IDC_HILITERECT, BM_GETCHECK, 0, 0) != BST_CHECKED);
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetHiliteRect(hiliteRect);
				}
				Update();
				return true;
			}
			return false;

		case IDC_CORNERX:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int right = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_CORNERX)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetRight(right);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_CORNERY:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int bottom = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_CORNERY)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetBottom(bottom);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_THICKNESS:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int thickness = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_THICKNESS)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetThickness(thickness);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_HRADIUS:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int hRadius = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_HRADIUS)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetHRadius(hRadius);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_VRADIUS:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int vRadius = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_VRADIUS)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetVRadius(vRadius);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_FILLNORMAL:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool fillNormal = (SendDlgItemMessage(hWnd(), IDC_FILLNORMAL, BM_GETCHECK, 0, 0) != BST_CHECKED);;
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetFillNormal(fillNormal);
				}
				Update();
				return true;
			}
			return false;

		case IDC_FILLHILITE:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool fillHilite = (SendDlgItemMessage(hWnd(), IDC_FILLHILITE, BM_GETCHECK, 0, 0) != BST_CHECKED);;
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetFillHilite(fillHilite);
				}
				Update();
				return true;
			}
			return false;

		case IDC_AUTORESIZE:
			if (HIWORD(wParam) == BN_CLICKED) {
				bool autoResize = (SendDlgItemMessage(hWnd(), IDC_AUTORESIZE, BM_GETCHECK, 0, 0) != BST_CHECKED);;
				for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
					m_Menu->GetMenuItem(index)->SetAutoResize(autoResize);
				}
				Update();
				return true;
			}
			return false;

		case IDC_HMARGIN:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int hMargin = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_HMARGIN)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetHMargin(hMargin);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_VMARGIN:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				hasChanged = false;
				return true;
			case EN_CHANGE:
				hasChanged = true;
				return true;
			case EN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int vMargin = atoi(String().fromWindow(GetDlgItem(hWnd(), IDC_VMARGIN)).toAnsi().c_str());
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetVMargin(vMargin);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_ALIGNMENT:
			switch (HIWORD(wParam)) {
			case CBN_SETFOCUS:
				hasChanged = false;
				return true;
			case CBN_SELCHANGE:
				hasChanged = true;
				return true;
			case CBN_KILLFOCUS:
				if (hasChanged) {
					hasChanged = false;
					int alignment = (int) SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_GETCURSEL, 0, 0);
					for (int index = 0; index < m_Menu->GetMenuItemCount(); index++) if (m_Menu->IsMenuItemSelected(index)) {
						m_Menu->GetMenuItem(index)->SetAlignment(alignment);
					}
					Update();
				}
				return true;
			}
			return false;

		case IDC_BACK:
			if (HIWORD(wParam) == BN_CLICKED) {
				SendMessage(hWnd(), WM_CLOSE, 0, 0);
				return true;
			}
			return false;
		}
		return false;

	case WM_MEASUREITEM:
		MeasureItem((LPMEASUREITEMSTRUCT) lParam);
		return true;

	case WM_DRAWITEM:
		DrawItem((LPDRAWITEMSTRUCT) lParam);
		return true;

	case WM_CLOSE:
		EndDialog(hWnd(), 0);
		return true;
	}
	return false;
}

int CALLBACK InitMenuItemFonts(LPENUMLOGFONT Font, LPNEWTEXTMETRIC Metric, int FontType, LPARAM lParam)
{
	HWND hWnd;
	hWnd = (HWND) lParam;
	SendDlgItemMessage(hWnd, IDC_FONTFACE, CB_ADDSTRING, 0, (LPARAM)(Font->elfLogFont.lfFaceName));
	return 1;
}

void MenuDialog::InitDialog()
{
	SetWindowLong(GetDlgItem(hWnd(), IDC_PREVIEW), GWL_USERDATA, (LONG) this);

	HDC dc = GetDC(hWnd());
	EnumFontFamilies(dc, NULL, (FONTENUMPROC) InitMenuItemFonts, (LPARAM) hWnd());
	ReleaseDC(hWnd(), dc);

	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Top-left"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Top-center"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Top-right"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Center-left"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Center"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Center-right"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Bottom-left"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Bottom-center"));
	SendDlgItemMessage(hWnd(), IDC_ALIGNMENT, CB_ADDSTRING, 0, (LPARAM) TEXT("Bottom-right"));

	UpdateBitmap();
	Update(true);
}

void MenuDialog::MeasureItem(LPMEASUREITEMSTRUCT mis)
{
	int index = (int) mis->itemData;
	StringVectorVector strings;
	if (m_Menu->IsMenuItemIndexValid(index)) {
		m_Menu->GetMenuItem(index)->GetDescription(strings);
	}
	MeasureListItem(hWnd(), mis, strings, 0);
}

void MenuDialog::DrawItem(LPDRAWITEMSTRUCT dis)
{
	int index = (int) dis->itemData;
	StringVectorVector strings;
	bool framed = false;
	if (m_Menu->IsMenuItemIndexValid(index)) {
		m_Menu->GetMenuItem(index)->GetDescription(strings);
		framed = (m_Menu->GetFirstSelected() == index);
	}
	DrawListItem(hWnd(), dis, strings, index + 1);
}

void MenuDialog::OnSelChange()
{
	for (int i = 0; i < m_Menu->GetMenuItemCount(); i++) {
		bool selected = (SendDlgItemMessage(hWnd(), IDC_MENUITEMLIST, LB_GETSEL, i, 0) != 0);
		m_Menu->GetMenuItem(i)->SetSelected(selected);
	}
	m_Menu->Recalculate();
	Update();
}

void MenuDialog::BrowseBackgroundImage()
{
	TCHAR fileName[MAX_PATH];
	lstrcpyn(fileName, m_Menu->GetBackgroundImage().c_str(), MAX_PATH);
	OPENFILENAME ofn;
	ZeroMemory(&ofn, sizeof(OPENFILENAME));
	ofn.lStructSize     = sizeof(OPENFILENAME);
	ofn.hwndOwner       = hWnd();
	ofn.hInstance       = hInstance;
	ofn.lpstrFilter     = TEXT("Image files\0")
		TEXT("*.bmp;*.cut;*.dcx;*.dds;*.ico;*.jpg;*.jpeg;*.lbm;*.lif;*.mdl;*.pcd;*.pcx;*.pic;*.png;*.pnm;")
		TEXT("*.psd;*.psp;*.raw;*.sgi;*.tga;*.tif;*.tiff;*.wal;*.act;*.pal;*.hdr\0") // from DevIL doc
		TEXT("All files\0*.*\0\0");
	ofn.lpstrFile       = fileName;
	ofn.nMaxFile        = MAX_PATH;
	ofn.lpstrInitialDir = m_Menu->GetGroup()->GetProject()->GetOpenFileDirectory().c_str();
	ofn.lpstrTitle      = TEXT("Browse background image");
	ofn.Flags           = OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_EXPLORER;
	if (GetOpenFileName(& ofn)) {
		m_Menu->SetBackgroundImage(String(fileName));
		m_Menu->GetGroup()->GetProject()->SetOpenFileDirectory(String(fileName).replaceFileName(S("")));
		UpdateBitmap();
		Update();
	}
}

void MenuDialog::ChooseMenuColor(int which)
{
	CHOOSECOLOR cc;
	cc.lStructSize    = sizeof(CHOOSECOLOR);
	cc.hwndOwner      = hWnd();
	cc.hInstance      = NULL;
	cc.lpCustColors   = CustomColors;
	cc.Flags          = CC_ANYCOLOR | CC_FULLOPEN | CC_RGBINIT;
	cc.lCustData      = 0;
	cc.lpfnHook       = NULL;
	cc.lpTemplateName = NULL;

	switch (which) {
	case MENUCOLOR_BACKGROUND:
		cc.rgbResult = m_Menu->GetBackgroundColor();
		break;
	case MENUCOLOR_NORMAL:
		cc.rgbResult = m_Menu->GetNormalColor();
		break;
	case MENUCOLOR_SELECT:
		cc.rgbResult = m_Menu->GetSelectColor();
		break;
	case MENUCOLOR_HILITE:
		cc.rgbResult = m_Menu->GetHiliteColor();
		break;
	}
	if (ChooseColor(&cc)) {
		switch (which) {
		case MENUCOLOR_BACKGROUND:
			m_Menu->SetBackgroundColor(cc.rgbResult);
			break;
		case MENUCOLOR_NORMAL:
			m_Menu->SetNormalColor(cc.rgbResult);
			break;
		case MENUCOLOR_SELECT:
			m_Menu->SetSelectColor(cc.rgbResult);
			break;
		case MENUCOLOR_HILITE:
			m_Menu->SetHiliteColor(cc.rgbResult);
			break;
		}
		Update();
	}
}

LRESULT MenuDialog::PreviewWndProc(HWND previewWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message)
	{
		case WM_PAINT: 
		{
			PAINTSTRUCT ps;
			BeginPaint(previewWnd, & ps);
			RECT clientRect;
			GetClientRect(previewWnd, & clientRect);
			int windowWidth  = clientRect.right - clientRect.left;
			int windowHeight = clientRect.bottom - clientRect.top;
			RECT pad;
			pad.left = pad.top = -100;
			pad.right = pad.bottom = 0;
			HDC dc = GetDC(previewWnd);
			DrawEdge(dc, & pad, EDGE_SUNKEN, BF_RECT | BF_ADJUST);
			pad.left += 100;
			pad.top  += 100;
			pad.right  *= -1;
			pad.bottom *= -1;
			int videoWidth  = m_Menu->GetGroup()->GetProject()->GetVideoWidth();
			int videoHeight = m_Menu->GetGroup()->GetProject()->GetVideoHeight();
			m_ImageRect.left   = (windowWidth - videoWidth - pad.left - pad.right) / 2;
			m_ImageRect.top    = (windowHeight - videoHeight - pad.top - pad.bottom) / 2;
			m_ImageRect.right  = m_ImageRect.left + pad.left + videoWidth  + pad.right;
			m_ImageRect.bottom = m_ImageRect.top  + pad.top  + videoHeight + pad.bottom;
			HBITMAP compBitmap = CreateCompatibleBitmap(dc, windowWidth, windowHeight);
			HDC compDC = CreateCompatibleDC(dc);
			SelectObject(compDC, compBitmap);
			SelectObject(compDC, GetStockObject(NULL_PEN));
			SelectObject(compDC, GetSysColorBrush(COLOR_3DFACE));
			Rectangle(compDC, 0, 0, windowWidth + 1, windowHeight + 1);
			DrawEdge(compDC, & m_ImageRect, EDGE_SUNKEN, BF_RECT | BF_ADJUST);
			if (m_Bitmap) {
				HDC imageDC = CreateCompatibleDC(dc);
				SelectObject(imageDC, m_Bitmap);
				BitBlt(compDC, m_ImageRect.left, m_ImageRect.top, videoWidth, videoHeight, imageDC, 0, 0, SRCCOPY);
				DeleteDC(imageDC);
			} else {
				HBRUSH brush = CreateSolidBrush(m_Menu->GetBackgroundColor());
				FillRect(compDC, & m_ImageRect, brush);
				DeleteObject(brush);
			}
			MenuDrawer::DrawMenuItems(m_Menu, compDC, m_ImageRect.left, m_ImageRect.top, true, 0, 0, false, true);
			BitBlt(dc, clientRect.left, clientRect.top, windowWidth, windowHeight, compDC, 0, 0, SRCCOPY);
			DeleteDC(compDC);
			DeleteObject(compBitmap);
			ReleaseDC(previewWnd, dc);
			EndPaint(previewWnd, & ps);
			return 0;
		}
		
		case WM_LBUTTONDOWN:
		{
			if (! m_IsDragging) {
				m_IsDragging = true;
				m_ReallyDragging = false;
				GetCursorPos(& m_DragOrigin);
				ScreenToClient(previewWnd, & m_DragOrigin);
				m_DragOrigin.x -= m_ImageRect.left;
				m_DragOrigin.y -= m_ImageRect.top;
				m_DragOp = DRAG_SELECT;
				for (int mindex = 0; mindex < m_Menu->GetMenuItemCount(); mindex++) {
					if (m_Menu->GetMenuItem(mindex)->IsSelected()) {
						for (int op = 0; (m_DragOp == DRAG_SELECT) && (op < DRAG_CORNERS); op++) {
							if (PtInRect(& m_Menu->GetMenuItem(mindex)->GetDragRect(op), m_DragOrigin)) {
								m_DragOp = op;
							}
						}
					}
				}
				if (m_DragOp == DRAG_SELECT) {
					for (int mindex = 0; mindex < m_Menu->GetMenuItemCount(); mindex++) {
						if (m_Menu->GetMenuItem(mindex)->IsSelected()) {
							if (PtInRect(& m_Menu->GetMenuItem(mindex)->GetBaseRect(), m_DragOrigin)) {
								m_DragOp = DRAG_MOVE;
							}
						}
					}
				}
				if (m_DragOp == DRAG_SELECT) {
					for (int mindex = 0; mindex < m_Menu->GetMenuItemCount(); mindex++) {
						m_Menu->GetMenuItem(mindex)->SetSelected(false);
						if ((m_DragOp == DRAG_SELECT) && PtInRect(& m_Menu->GetMenuItem(mindex)->GetBaseRect(), m_DragOrigin)) {
							m_DragOp = DRAG_MOVE;
							m_Menu->GetMenuItem(mindex)->SetSelected(true);
						}
					}
				}
				m_OrigXY.clear();
				for (int mindex = 0; mindex < m_Menu->GetMenuItemCount(); mindex++) {
					MenuItem * menuItem = m_Menu->GetMenuItem(mindex);
					POINT pt;
					pt.x = menuItem->GetX();
					pt.y = menuItem->GetY();
					switch (m_DragOp) {
					case DRAG_TOPRIGHT:
					case DRAG_RIGHT:
					case DRAG_BOTTOMRIGHT:
						pt.x = menuItem->GetRight();
					}
					switch (m_DragOp) {
					case DRAG_BOTTOMLEFT:
					case DRAG_BOTTOM:
					case DRAG_BOTTOMRIGHT:
						pt.y = menuItem->GetBottom();
					}
					m_OrigXY.push_back(pt);
				}
				switch (m_DragOp) {
				case DRAG_SELECT:
					SetCursor(LoadCursor(NULL, IDC_CROSS));
					break;
				case DRAG_MOVE:
					SetCursor(LoadCursor(NULL, IDC_SIZEALL));
					break;
				case DRAG_TOPLEFT:
				case DRAG_BOTTOMRIGHT:
					SetCursor(LoadCursor(NULL, IDC_SIZENWSE));
					break;
				case DRAG_TOP:
				case DRAG_BOTTOM:
					SetCursor(LoadCursor(NULL, IDC_SIZENS));
					break;
				case DRAG_TOPRIGHT:
				case DRAG_BOTTOMLEFT:
					SetCursor(LoadCursor(NULL, IDC_SIZENESW));
					break;
				case DRAG_LEFT:
				case DRAG_RIGHT:
					SetCursor(LoadCursor(NULL, IDC_SIZEWE));
					break;
				}
				SetCapture(previewWnd);
				Update();
			}
			return 0;
		}
		
		case WM_MOUSEMOVE:
		{
			if (m_IsDragging) {
				POINT pt;
				GetCursorPos(& pt);
				ScreenToClient(previewWnd, & pt);
				pt.x -= m_ImageRect.left;
				pt.y -= m_ImageRect.top;
				int dx = pt.x - m_DragOrigin.x;
				int dy = pt.y - m_DragOrigin.y;
				if (! m_ReallyDragging) {
					int dragx = GetSystemMetrics(SM_CXDRAG);
					int dragy = GetSystemMetrics(SM_CYDRAG);
					m_ReallyDragging = (dx < -dragx || dragx < dx || dy < -dragy || dragy < dy);
				}
				if (m_ReallyDragging) {
					for (int mindex = 0; mindex < m_Menu->GetMenuItemCount(); mindex++) {
						MenuItem * menuItem = m_Menu->GetMenuItem(mindex);
						if (m_DragOp == DRAG_SELECT) { 
							if (PtInRect(& menuItem->GetDrawnRect(), pt)) {
								menuItem->SetSelected(true);
							}
						} else if (menuItem->IsSelected()) {
							int videoWidth  = m_Menu->GetGroup()->GetProject()->GetVideoWidth()  - 1;
							int videoHeight = m_Menu->GetGroup()->GetProject()->GetVideoHeight() - 1;
							if (m_DragOp == DRAG_MOVE) {
								int x = m_OrigXY[mindex].x + dx;
								int y = m_OrigXY[mindex].y + dy;
								int r = m_OrigXY[mindex].x + menuItem->GetRight() - menuItem->GetX() + dx;
								int b = m_OrigXY[mindex].y + menuItem->GetBottom() - menuItem->GetY() + dy;
								if (x < 0) {
									r -= x;
									x = 0;
								} else if (r > videoWidth) {
									x -= (r - videoWidth);
									r = videoWidth;
								}
								if (y < 0) {
									b -= y;
									y = 0;
								} else if (b > videoHeight) {
									y -= (b - videoHeight);
									b = videoHeight;
								}
								menuItem->SetX(x, false);
								menuItem->SetY(y, false);
								menuItem->SetRight(r, false);
								menuItem->SetBottom(b, false);
							} else {
								switch (m_DragOp) {
								case DRAG_TOPLEFT: 
								case DRAG_LEFT:
								case DRAG_BOTTOMLEFT:
									menuItem->SetAutoResize(false, false);
									int x = m_OrigXY[mindex].x + dx;
									if (x > menuItem->GetRight()) {
										x = menuItem->GetRight();
									} else if (x < 0) {
										x = 0;
									}
									menuItem->SetX(x, false);
								}
								switch (m_DragOp) {
								case DRAG_TOPLEFT:
								case DRAG_TOP:
								case DRAG_TOPRIGHT:
									int y = m_OrigXY[mindex].y + dy;
									menuItem->SetAutoResize(false, false);
									if (y > menuItem->GetBottom()) {
										y = menuItem->GetBottom();
									} else if (y < 0) {
										y = 0;
									}
									menuItem->SetY(y, false);
								}
								switch (m_DragOp) {
								case DRAG_TOPRIGHT:
								case DRAG_RIGHT:
								case DRAG_BOTTOMRIGHT:
									menuItem->SetAutoResize(false, false);
									int r = m_OrigXY[mindex].x + dx;
									if (r < menuItem->GetX()) {
										r = menuItem->GetX();
									} else if (r > videoWidth) {
										r = videoWidth;
									}
									menuItem->SetRight(r, false);
								}
								switch (m_DragOp) {
								case DRAG_BOTTOMLEFT:
								case DRAG_BOTTOM:
								case DRAG_BOTTOMRIGHT:
									menuItem->SetAutoResize(false, false);
									int b = m_OrigXY[mindex].y + dy;
									if (b < menuItem->GetY()) {
										b = menuItem->GetY();
									} else if (b > videoHeight) {
										b = videoHeight;
									}
									menuItem->SetBottom(b, false);
								}
							}
							menuItem->Recalculate();
						}
					}
					Update();
				}
			}
			return 0;
		}
		
		case WM_LBUTTONUP:
		{
			if (m_IsDragging) {
				if (m_ReallyDragging) {
					m_Menu->SetChanged();
					m_ReallyDragging = false;
				}
				ReleaseCapture();
				SetCursor(LoadCursor(NULL, IDC_ARROW));
				m_IsDragging = false;
				Update();
			}
			return 0;
		}
		
	}

	return DefWindowProc(previewWnd, message, wParam, lParam);
}

LRESULT CALLBACK MenuColorControlProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	if (message == WM_PAINT) {
		HDC dc = GetDC(hWnd);
		PAINTSTRUCT ps;
		BeginPaint(hWnd, & ps);
		RECT m_ImageRect;
		GetClientRect(hWnd, & m_ImageRect);
		DrawEdge(dc, & m_ImageRect, EDGE_SUNKEN, BF_RECT | BF_ADJUST);
		HBRUSH brush = CreateSolidBrush(
			(IsWindowEnabled(hWnd) ? (COLORREF) GetWindowLong(hWnd, GWL_USERDATA) : GetSysColor(COLOR_3DFACE)));
		FillRect(dc, & m_ImageRect, brush);
		DeleteObject(brush);
		EndPaint(hWnd, & ps);
		ReleaseDC(hWnd, dc);
		return 0;
	} else {
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
}

LRESULT CALLBACK MenuPreviewControlProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	MenuDialog * menuDialog = (MenuDialog *) GetWindowLong(hWnd, GWL_USERDATA);
	if (menuDialog) {
		return menuDialog->PreviewWndProc(hWnd, message, wParam, lParam);
	} else {
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
}

void RegisterMenuDialogControls()
{
	WNDCLASS wcl;
	wcl.style         = 0;
	wcl.cbClsExtra    = 0;
	wcl.cbWndExtra    = 0;
	wcl.hInstance     = hInstance;
	wcl.hIcon         = NULL;
	wcl.hCursor       = LoadCursor(NULL, IDC_ARROW);
	wcl.hbrBackground = (HBRUSH) GetStockObject(HOLLOW_BRUSH);
	wcl.lpszMenuName  = NULL;

	wcl.lpfnWndProc   = MenuColorControlProc;
	wcl.lpszClassName = TEXT("MenuColor");
	RegisterClass(& wcl);

	wcl.lpfnWndProc   = MenuPreviewControlProc;
	wcl.lpszClassName = TEXT("MenuPreview");
	RegisterClass(& wcl);
}
