#include "menudrawer.h"
#include "group.h"
#include "menu.h"
#include "menuitem.h"
#include "tools.h"
#include "config.h"
#include "multiplexer.h"

#include "EasyBMP/EasyBMP.h"
#include "EasyBMP/EasyBMP_win32.h"
#include "winutils/winutils.h"

MenuDrawer::MenuDrawer(ProjectPtr project, int groupIndex, int whichMenu) : Thread()
{
	m_Project = project;
	m_GroupIndex = groupIndex;
	m_WhichMenu = whichMenu;
}

MenuDrawer::~MenuDrawer()
{
}

String MenuDrawer::GetDescription() const
{
	String descr = S("Creating ");
	switch (m_WhichMenu) {
	case MENU_ROOT:
		descr += S("root");
		break;
	case MENU_COMBINED:
		descr += S("audio + subtitle");
		break;
	case MENU_AUDIO:
		descr += S("audio");
		break;
	case MENU_SUBTITLE:
		descr += S("subtitle");
		break;
	}
	descr += (S(" menu of group ") + String(m_GroupIndex + 1));
	return descr;
}

void MenuDrawer::DrawMenuItems(Menu * menu, HDC dc, int xoffset, int yoffset, 
	bool preview, COLORREF bgColor, COLORREF fgColor, bool highlighted, bool dragBoxes)
{
	double aspectCorr = menu->GetGroup()->GetProject()->GetAspectCorr(menu->GetGroup()->GetMenuAspectRatio());
	for (int index = 0; index < menu->GetMenuItemCount(); index++) {
		MenuItem * menuItem = menu->GetMenuItem(index);
		const RECT & baseRect = menuItem->GetBaseRect();
		const RECT & textRect = menuItem->GetTextRect();
		if (preview) {
			highlighted = menuItem->IsSelected();
			fgColor = menu->GetColor(highlighted ? MENUCOLOR_HILITE : MENUCOLOR_NORMAL);
			bgColor = GetPixel(dc, (textRect.left + textRect.right) / 2, (textRect.top + textRect.bottom) / 2);
		}
		bool filled = false;
		if (menuItem->GetRectangle()) {
			if ((menuItem->GetHiliteRect() && highlighted) || (menuItem->GetNormalRect() && ! highlighted)) {
				int bheight = menuItem->GetThickness();
				int bwidth  = (int) (0.5 + bheight * aspectCorr);
				int hradius = menuItem->GetHRadius() * 2;
				int vradius = menuItem->GetVRadius() * 2;
				filled = ((menuItem->GetFillHilite() && highlighted) || (menuItem->GetFillNormal() && ! highlighted));
				HPEN pen = CreatePen(PS_SOLID, 1, fgColor);
				HGDIOBJ origPen = SelectObject(dc, pen);
				HBRUSH brush = (HBRUSH) (filled ? CreateSolidBrush(fgColor) : GetStockObject(HOLLOW_BRUSH));
				HGDIOBJ origBrush = SelectObject(dc, brush);
				if (filled) {
					RoundRect(dc, 
						xoffset + baseRect.left - bwidth, yoffset + baseRect.top - bwidth, 
						xoffset + baseRect.right + bwidth, yoffset + baseRect.bottom + bwidth, 
						hradius, vradius);
				} else {
					for (int dx = 1; dx <= bwidth; dx++) {
						for (int dy = 1; dy <= bheight; dy++) {
							RoundRect(dc, 
								xoffset + baseRect.left - dx, yoffset + baseRect.top - dy, 
								xoffset + baseRect.right + dx, yoffset + baseRect.bottom + dy, 
								hradius, vradius);
						}
					}
				}
				SelectObject(dc, origBrush);
				if (filled) DeleteObject(brush);
				SelectObject(dc, origPen);
				DeleteObject(pen);
			}
		}
		if (menuItem->GetCaption().length()) {
			if ((menuItem->GetHiliteText() && highlighted) || (menuItem->GetNormalText() && ! highlighted)) {
				HFONT font = menu->GetMenuItem(index)->CreateFont();
				HGDIOBJ origFont = SelectObject(dc, font);
				COLORREF origColor = SetTextColor(dc, (filled ? bgColor : fgColor));
				int origMode = SetBkMode(dc, TRANSPARENT);
				TextOut(dc, xoffset + textRect.left, yoffset + textRect.top,
					menu->GetMenuItem(index)->GetCaption().c_str(),
					(int) menu->GetMenuItem(index)->GetCaption().length());
				SetBkMode(dc, origMode);
				SetTextColor(dc, origColor);
				SelectObject(dc, origFont);
				DeleteObject(font);
			}
		}
	}
	if (dragBoxes) {
		HPEN origPen = (HPEN) SelectObject(dc, GetStockObject(WHITE_PEN));
		HBRUSH origBrush = (HBRUSH) SelectObject(dc, GetStockObject(HOLLOW_BRUSH));
		int videoWidth  = menu->GetGroup()->GetProject()->GetVideoWidth();
		int videoHeight = menu->GetGroup()->GetProject()->GetVideoHeight();
		RECT safeRect;
		safeRect.left   = xoffset + videoWidth  / 10;
		safeRect.top    = yoffset + videoHeight / 10;
		safeRect.right  = xoffset + videoWidth  * 9 / 10;
		safeRect.bottom = yoffset + videoHeight * 9 / 10;
		Rectangle(dc, safeRect.left, safeRect.top, safeRect.right, safeRect.bottom);
		DrawFocusRect(dc, & safeRect);
		SelectObject(dc, GetStockObject(BLACK_BRUSH));
		for (int index = 0; index < menu->GetMenuItemCount(); index++) {
			MenuItem * menuItem = menu->GetMenuItem(index);
			if (menuItem->IsSelected()) {
				for (int index = 0; index < DRAG_CORNERS; index++) {
					const RECT & rect = menuItem->GetDragRect(index);
					Rectangle(dc, xoffset + rect.left, yoffset + rect.top, xoffset + rect.right, yoffset + rect.bottom);
				}
			}
		}
		SelectObject(dc, origBrush);
		SelectObject(dc, origPen);
	}
}

bool MenuDrawer::Draw(Menu * menu, COLORREF bgColor, int menuColor, const String & fileName, LPRECT videoRect)
{
	HDC dc = CreateCompatibleDC(NULL);
	HBITMAP bitmap = CreateBitmap(m_Project->GetVideoWidth(), m_Project->GetVideoHeight(), 1, 32, NULL);
	SelectObject(dc, bitmap);
	FillRect(dc, videoRect, (HBRUSH) GetStockObject(BLACK_BRUSH));
	DrawMenuItems(menu, dc, 0, 0, false, 0, 255, (menuColor != MENUCOLOR_NORMAL));
	BMP image;
	image.SetSize(m_Project->GetVideoWidth(), m_Project->GetVideoHeight());
	image.SetBitDepth(4);
	COLORREF fgColor = menu->GetColor(menuColor);
	RGBApixel color[16];
	for (int c = 0; c < 16; c++) {
		if (c == 1) {
			color[c].Red   = GetRValue(fgColor);
			color[c].Green = GetGValue(fgColor);
			color[c].Blue  = GetBValue(fgColor);
			color[c].Alpha = 0;
		} else {
			color[c].Red   = GetRValue(bgColor);
			color[c].Green = GetGValue(bgColor);
			color[c].Blue  = GetBValue(bgColor);
			color[c].Alpha = 255;
		}
		image.SetColor(c, color[c]);
	}
	for (int x = 0; x < m_Project->GetVideoWidth(); x++) {
		for (int y = 0; y < m_Project->GetVideoHeight(); y++) {
			image.SetPixel(x, y, color[GetPixel(dc, x, y) / 128]);
		}
	}
	DeleteObject(bitmap);
	DeleteDC(dc);
	return image.WriteToFile(fileName.toAnsi().c_str());

	/*
	HDC dc = CreateCompatibleDC(NULL);
	HBITMAP bitmap = CreateBitmap(m_Project->GetVideoWidth(), m_Project->GetVideoHeight(), 1, 32, NULL);
	SelectObject(dc, bitmap);
	HBRUSH brush = CreateSolidBrush(bgColor);
	FillRect(dc, videoRect, brush);
	DeleteObject(brush);
	COLORREF fgColor = menu->GetColor(menuColor);
	DrawMenuItems(menu, dc, 0, 0, false, bgColor, fgColor, (menuColor != MENUCOLOR_NORMAL));
	BMP image;
	HBITMAPtoBMP(dc, bitmap, image);
	DeleteObject(bitmap);
	DeleteDC(dc);
	image.SetBitDepth(4);
	RGBApixel color[16];
	for (int c = 0; c < 16; c++) {
		if (c == 1) {
			color[c].Red   = GetRValue(fgColor);
			color[c].Green = GetGValue(fgColor);
			color[c].Blue  = GetBValue(fgColor);
			color[c].Alpha = 0;
		} else {
			color[c].Red   = GetRValue(bgColor);
			color[c].Green = GetGValue(bgColor);
			color[c].Blue  = GetBValue(bgColor);
			color[c].Alpha = 255;
		}
		image.SetColor(c, color[c]);
	}
	return image.WriteToFile(fileName.toAnsi().c_str());
	*/
}

bool MenuDrawer::OnRun()
{
	bool ok = true;

	String baseFileName = String(m_Project->GetOutputDirectory()
		+ S("\\set") + String(m_GroupIndex + 1)
		+ S("menu") + String(m_WhichMenu)).toFullPath();
	
	GroupPtr group = m_Project->GetGroup(m_GroupIndex);
	MenuPtr menu;
	switch (m_WhichMenu) {
	case MENU_ROOT:
		menu = group->GetMenu();
		break;
	case MENU_COMBINED:
		menu = group->GetCombinedMenu();
		break;
	case MENU_AUDIO:
		menu = group->GetAudioMenu();
		break;
	case MENU_SUBTITLE:
		menu = group->GetSubtitleMenu();
		break;
	}
	
	RECT videoRect; 
	videoRect.left   = 0;
	videoRect.top    = 0;
	videoRect.right  = m_Project->GetVideoWidth();
	videoRect.bottom = m_Project->GetVideoHeight();

	COLORREF bgColor = menu->GetBackgroundColor();
	String nmFileName = baseFileName + S("nm.bmp");
	Draw(menu, bgColor, MENUCOLOR_NORMAL, nmFileName, & videoRect);
	String hlFileName = baseFileName + S("hl.bmp");
	Draw(menu, bgColor, MENUCOLOR_HILITE, hlFileName, & videoRect);
	String slFileName = baseFileName + S("sl.bmp");
	Draw(menu, bgColor, MENUCOLOR_SELECT, slFileName, & videoRect);

	String nmPngFileName = baseFileName + S("nm.png");
	AddFileToDelete(nmPngFileName);
	String hlPngFileName = baseFileName + S("hl.png");
	AddFileToDelete(hlPngFileName);
	String slPngFileName = baseFileName + S("sl.png");
	AddFileToDelete(slPngFileName);
	String avsFileName = baseFileName + S(".avs");
	AddFileToDelete(avsFileName);
	String m2vFileName = baseFileName + S(".m2v");
	AddFileToDelete(m2vFileName);
	String tmpFileName = baseFileName + S(".tmp");
	AddFileToDelete(tmpFileName);
	String xmlFileName = baseFileName + S(".xml");
	AddFileToDelete(xmlFileName);
	String mpgFileName = baseFileName + S(".mpg");
	String logFileName = String(m_Project->GetOutputDirectory() + S("\\") + S("log.txt")).toFullPath();

	int frames = (int) (0.5 + m_Project->GetFrameRate());
	double length = 1.0 * frames / m_Project->GetFrameRate();

	FILE * f = fopen(avsFileName.toAnsi().c_str(), "wt");
	if (f) {
		if (menu->GetBackgroundImage().length()) {
			fprintf(f, "ImageSource(%s, 0, %d)\n", menu->GetBackgroundImage().enclose().toAnsi().c_str(), frames - 1);
			fprintf(f, "Lanczos4Resize(%d, %d)\n", m_Project->GetVideoWidth(), m_Project->GetVideoHeight());
			fprintf(f, "ConvertToYV12()\n");
		} else {
			fprintf(f, "BlankClip(length=%d, width=%d, height=%d, pixel_type=\"YV12\", color=%d)\n",
				frames, m_Project->GetVideoWidth(), m_Project->GetVideoHeight(), bgColor);
		}
		fprintf(f, "AssumeFPS(\"%s\")\n", (m_Project->GetDVDFormat() == DVDFORMAT_PAL ? "pal_video" : "ntsc_video"));
		fclose(f);
	}

	String encoderExe;
	String encoderCommand;
	if (g_Config->GetEncoder() == ENCODER_HC) {
		encoderExe = tools.HCenc;
		String iniFileName = baseFileName + S(".ini");
		AddFileToDelete(iniFileName);
		String dbsFileName = m_Project->GetOutputDirectory();
		encoderCommand = S("-ini ") + iniFileName.enclose();
		f = fopen(iniFileName.toAnsi().c_str(), "wt");
		if (f) {
			fprintf(f, "*INFILE  %s\n", avsFileName.toAnsi().c_str());
			fprintf(f, "*OUTFILE %s\n", m2vFileName.toAnsi().c_str());
			fprintf(f, "*DBPATH  %s\n", dbsFileName.toAnsi().c_str());
			fprintf(f, "*LOGFILE %s\n", logFileName.toAnsi().c_str());
			fprintf(f, "*ASPECT %s\n", menu->GetGroup()->GetMenuAspectRatioStr().toAnsi().c_str());
			fprintf(f, "*BITRATE %d\n", m_Project->GetVideoBitRate());
			fprintf(f, "*MAXBITRATE %d\n", m_Project->GetMaxVideoBitRate());
			fprintf(f, "*PROFILE BEST\n");
			fprintf(f, "*WAIT 0\n");
			if (g_Config->GetHideEncoder()) {
				fprintf(f, "*SILENT\n");
			}
			fclose(f);
		}
	} else { // ENCODER_QU
		encoderExe = tools.QuEnc;
		encoderCommand = S("-auto -close")
			+ S(" -i ") + avsFileName.enclose() 
			+ S(" -o ") + m2vFileName.enclose()
			+ S(" -b ") + String(m_Project->GetVideoBitRate()) 
			+ S(" -aspectratio ") + menu->GetGroup()->GetMenuAspectRatioStr()
			+ S(" -maxbitrate 9600 -2 -hq -mpeg2mux nomux");
		if (g_Config->GetHideEncoder()) encoderCommand += S(" -silent");
	}

	f = fopen(xmlFileName.toAnsi().c_str(), "wt");
	if (f) {
		fprintf(f, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
		fprintf(f, "<subpictures>\n<stream>\n");
		fprintf(f, "<spu image=\"%s\" highlight=\"%s\" select=\"%s\"\n"
			" start=\"0:00:00\" transparent=\"%02X%02X%02X\" force=\"yes\">\n",
			nmPngFileName.xmlenc().toUtf8().c_str(),
			hlPngFileName.xmlenc().toUtf8().c_str(),
			slPngFileName.xmlenc().toUtf8().c_str(),
			GetRValue(bgColor), GetGValue(bgColor), GetBValue(bgColor));
		for (int index = 0; index < menu->GetMenuItemCount(); index++) {
			const RECT & rect = menu->GetMenuItem(index)->GetAdjustedRect();
			fprintf(f, "  <button name=\"b%d\" x0=\"%d\" y0=\"%d\" x1=\"%d\" y1=\"%d\"/>\n",
				index, rect.left, rect.top, rect.right, rect.bottom);
		}
		fprintf(f, "</spu>\n</stream>\n</subpictures>\n");
		fclose(f);
	}

	String bmp2pngCommand = S("-X -Q -E ") + nmFileName.enclose() + S(" ") + hlFileName.enclose() + S(" ") + slFileName.enclose();
	String mplexCommand   = S("-f 8 -o ") + tmpFileName.enclose() 
		+ S(" ") + m2vFileName.enclose() + S(" ") + tools.silent.enclose();
	String spumuxCommand  = S("-s 0 ") + xmlFileName.enclose();

	ok = ok && ExecuteCommand(tools.bmp2png.c_str(), bmp2pngCommand.c_str(),
		NULL, NULL, logFileName.c_str(), isCancelled, waitingForUser);
	ExecuteCommand(encoderExe.c_str(), encoderCommand.c_str(), 
		NULL, NULL, NULL, isCancelled, waitingForUser);
	if (ok) {
		ok = ok && ExecuteCommand(tools.mplex.c_str(), (LPTSTR) mplexCommand.c_str(), 
			NULL, NULL, logFileName.c_str(), isCancelled, waitingForUser);
		if (! ok) ok = Multiplexer::IsFakeError(logFileName);
	}
	ok = ok && ExecuteCommand(tools.spumux.c_str(), (LPTSTR) spumuxCommand.c_str(),
		tmpFileName.c_str(), mpgFileName.c_str(), logFileName.c_str(), isCancelled, waitingForUser);

	if (ok) DeleteFilesNow();
	return ok;
}
